WIKIDATA_ITS_FAN_V1 = "https://www.wikidata.org/entity/Q131549102"
WIKIDATA_FAN_OPERATING_POINT = "https://www.wikidata.org/entity/Q137525225"

WIKIDATA_AFFINITY_LAWS = "https://www.wikidata.org/entity/Q632736"
